// Python compatibility functions
var Py = {
  zip : rows=>rows[0].map((_,c)=>rows.map(row=>row[c])),
  get : function(obj,key,dflt) {return key in obj ? obj[key] : dflt;}
};

function FiringHistory() {
  this.factTupleToRules = {};
  this.factsToFactTuples = {};
} 

FiringHistory.prototype = {
    addFiring : function(facts, ruleID) {
        var factT = JSON.stringify(facts);
        var ruleD = Py.get( this.factTupleToRules, factT, {} );
        ruleD[ ruleID ] = 1;
        this.factTupleToRules[ factT ] = ruleD;
        for (let fact of facts) {
            var factTL = Py.get( this.factsToFactTuples, fact, [] );
            if ( !factTL.includes(factT) ) {
                factTL.push( factT );
                this.factsToFactTuples[ fact ] = factTL;
            }
        }
    },
    handleFactDeletion : function(fact) {
        var factTL = Py.get( this.factsToFactTuples, fact, [] );
        for (let factT of factTL ) {
            if ( this.factTupleToRules[ factT ] )
                delete this.factTupleToRules[ factT ];
        }
        if (factTL)
            delete this.factsToFactTuples[ fact ];
    },
    alreadyFired : function(facts, ruleID) {
        var factT = JSON.stringify( facts );
        if (! this.factTupleToRules[ factT ] )
            return false;
        var ruleD = this.factTupleToRules[ factT ];
        return ruleD[ ruleID ] !== undefined;
    },
    dump : function() {
        console.log( 'Facts to Fact Tuples' );
        for (let fac of this.factsToFactTuples) {
            console.log( fac );
            for (let tup of this.factsToFactTuples[fac])
                console.log( '  ',tup );
        }
        console.log( 'Fact Tuples to Rules' );
        for (let facT of this.factTupleToRules) {
            console.log( facT );
            for (let rul of this.factTupleToRules[facT])
                console.log( '  ',rul );
        }
    }
};

        

function Rule(name) {
    this.name = name;
    this.patterns = [];
    this.neg = [];
    this.asserts = [];
    this.retracts = [];
    this.gensyms = [];
    this.id = Rule.prototype.nextRuleID ++;
}

Rule.prototype = {
    nextGensymInt : 1,
    nextRuleID : 1,
    addPattern : function(pat) {
        this.patterns.push(pat);
        this.neg.push(false);
    },        
    addPatternNot : function(pat) {
        this.patterns.push(pat);
        this.neg.push(true);
    },    
    addAssert : function(pat) {
        this.asserts.push(pat);
    },
    addRetract : function(pat) {
        this.retracts.push(pat);
    },
    addGensym : function(vbl) {
        this.gensyms.push(vbl);
    },
    fire : function(dict, cs) {
        //"Note: retracts ignored if not present."
        for (let vbl of this.gensyms) {
            var value = 'gen-' + Rule.prototype.nextGensymInt;
            Rule.prototype.nextGensymInt += 1;
            dict[vbl] = value;
        }
        for (let pat of this.retracts)
            cs.delFact( match.substitute(dict, pat) );
        for (let pat of this.asserts)
            cs.addFact( match.substitute(dict, pat) );
    },

    pprint : function() {
        var str = "";
        str += this.name + ": ";
        var neg = this.neg;
        if (!neg) {
            neg = [];
            for (let p of this.patterns)  neg.push("false");
        }
        for (let [p,n] of Py.zip([this.patterns, this.neg])) {
            if (n)
                str += '-';
            str += p[0] + '(';
            for (let a of p.slice(1,-1))
                str += a + " ";
            str += p[p.length-1];
            str += ') ';
        }
        str += '=> ';
        if (this.asserts) {
            for (let action of this.asserts) {
                str += action[0] + "(";
                for (let x of action.slice(1,-1))
                    str += x + " ";
                str += action[action.length-1];
                str += ') ';
            }
        }
        if (this.retracts) {
            for (let action of this.retracts) {
                str += '-';
                str += action[0] + "(";
                for (let x of action.slice(1,-1))
                    str += x + " ";
                str += action[action.length-1];
                str += ') ';
            }
        }
        str += '.';
        console.log(str);
    }
};
               
function equalSA(sa1,sa2) {
    if (sa1.length != sa2.length)  return false;
    for (var i=0;i<sa1.length;i++)
      if (sa1[i] != sa2[i])  return false;
    return true;
}
  

function Pyps( repeatFireB, watchB ) {
    repeatFireB = repeatFireB || false;
    watchB = watchB || false;
    this.memory = []; // Should probably be an object to facilitate deletion, and avoid multiple copies of the same fact
    this.ruleL = [];
    this.watchB = watchB;
    this.repeatFireB = repeatFireB;
    this.firingHistory = new FiringHistory();
}

Pyps.prototype = {

    addFact : function(fact) {
        // Do not add duplicate facts
        for (let f of this.memory)
          if (equalSA(fact,f))  return;
        this.memory.push( fact );
    },

    delFact : function(fact) { // Removes first copy only 
        function sameFact(f1,f2) {
            if (f1.length != f2.length)
                return false;
            for (var i=0;i<f1.length;i++) {
                if (f1[i]!=f2[i])  
                    return false;
            }
            return true;
        }
        for (var i=0;i<this.memory.length;i++) {
            if ( sameFact(fact,this.memory[i]) ) {
              this.firingHistory.handleFactDeletion( fact );
              this.memory.splice(i,1); // Delete i'th element and shift down
              return;
            }
        }
    },
    
    addRule : function(rule) {
        this.ruleL.push(rule);
    },

    facts : function() {
        var i = 0;
        for (let f of this.memory) {
            console.log( ''+i+': ('+f.join(" ")+')' );
            i = i + 1;
        }
    },

    agenda : function(verbose) {
        verbose = verbose || false;
        if (verbose)
          console.log( "Agenda" );
        var agendaL = [];
        for (let r of this.ruleL) {
            var oto = false;
            var iterator = match.matchSubset(r.patterns, this.memory, r.neg);
            while (true) {
                var next = iterator.next();
                if (next.done) break;
                [bd, facts] = next.value;
                //console.log("bd",bd,"facts",facts);
                if (bd==null)  break;
                if ( !this.repeatFireB ) {
                    if ( this.firingHistory.alreadyFired( facts, r.name ) )
                        continue;
                }
                agendaL.push( [r,bd,facts] );
                if (verbose)
                    console.log( r.name );
            }
        }
        return agendaL;
    },
                
    fire : function(r, bd, facts) {  
        this.firingHistory.addFiring(facts, r.name);
        r.fire(bd, this);
    },

    // should call agenda
    run1 : function() {
        //self.facts()
        var agendaL = this.agenda(this.watchB);
        if (agendaL.length==0)
            return false;
        [r,bd,facts] = agendaL[0];
        if (this.watchB)
            console.log( '----Firing rule', r.name );
        this.fire( r, bd, facts );
        return true;
    },

    run : function() {
        while( this.run1() ) {
            if (this.watchB) {
                console.log( 'Facts' );
                this.facts();
            }
        }
    },

    rules : function() {
        for (let r of this.ruleL)
            r.pprint();
    }
};

function *lexer(text) {
    /*
      The regexp in English:
        A global search for...
        A symbol (name of variable, constant, or predicate) OR
        An implication arrow "->" OR
        A single parenthesis, period, or minus sign
    */
    var tokenRegexp = /[?]*[A-Za-z_0-9]+|\=>|[()\.-]/g;
    var match;
    while ((match = tokenRegexp.exec(text)) !== null) {
        yield match[0];
    }
}

function isName(str) {
  return /^[A-Za-z_0-9]+$/.test(str);
}

function parser(tokens) {
  var current, done;
  
  function next() {
    var next = tokens.next();
    current = next.value;
    done = next.done;
  };
  
  function parseRule() {
    var bound_vars = {};
    if (!isName(current))
      throw new SyntaxError( "Expected rule name, got " + current );
    var atom, is_negative, varD;
    var r = new Rule(current);
    // Left-hand side
    while (true) {
      next();
      [atom, is_negative, varD] = parseAtom();
      if (!is_negative) {
        for (var v in varD)
          bound_vars[v] = true;
      }
      if (current == "=>")  break;
      if (is_negative)  r.addPatternNot(atom);
      else  r.addPattern(atom);
    }
    // Right-hand side
    while (current != ".") {
      next();
      [atom, is_negative, varD] = parseAtom();
      for (var v in varD) {
        if ( !bound_vars[v] )  {
          if (is_negative)
            throw new SyntaxError( "Unbound variable "+v+" in right-hand-side negative atom" );
          r.addGensym(v);
        }
        bound_vars[v] = true;
      }
      if (current == ".")  break;
      if (is_negative)  r.addRetract(atom);
      else  r.addAssert(atom);
    }
    return r;
  };
  
  function parseAtom() {
    if (current == "=>" || current == ".")
      return [null, null, {}];
      
    var atom = [];
    var varD = {};
      
    var is_negative = false;
    if (current == '-') {
      is_negative = true;
      next();
    }
    
    if (current===undefined)
      throw new SyntaxError( "Expected an atom, got undefined. (Missing period?)" );
    if (!isName(current))
      throw new SyntaxError( "Expected predicate, got " + current );
    atom.push( current );
    next();
    if (current != '(')
      throw new SyntaxError( "Expected (, got " + current );
    next();
    while (current != ')') {
      atom.push(current);
      var is_variable = false;
      if (current[0]=='?') {
        is_variable = true;
        varD[current] = true;
        current = current.slice(1);
      }
      if (!isName(current))
        throw new SyntaxError( "Expected constant or variable name, got " + current );
      next();
    }
    return [atom, is_negative, varD];
  };
  
  return {
    parseRules : function() {
      var rules = [];
      while (true) {
        next();
        if (done) break;
        rules.push( parseRule() );
      }
      return rules;
    },
    parseFacts : function() {
      var facts = [];
      while (true) {
        next();
        if (done) break;
        var [f,_,_] = parseAtom();
        facts.push(f);
      }
      return facts;
    }
  };
}
